@@ -28,9 +28,9 @@
 #' process one future at the time (`workers = 1L`), whereas HPC backends,
 #' where futures are resolved via separate jobs on a scheduler, can have
 #' multiple workers.  In the latter, the default is `workers = NULL`, which
-#' will resolve to `getOption("future.batchtools.workers")`.  If that is not
-#' specified, the value of environment variable `R_FUTURE_BATCHTOOLS_WORKERS`
-#' will be used.  If neither are specified, then the default is `100`.
+#' will resolve to
+#' \code{getOption("\link{future.batchtools.workers}")}.
+#' If neither are specified, then the default is `100`.
 #'
 #' @param finalize If TRUE, any underlying registries are
 #' deleted when this object is garbage collected, otherwise not.
@@ -62,6 +62,7 @@ BatchtoolsFuture <- function(expr = NULL, envir = parent.frame(),
                              registry = list(),
                              ...) {
   if (substitute) expr <- substitute(expr)
+  assert_no_positional_args_but_first()
 
   ## Record globals
   gp <- getGlobalsAndPackages(expr, envir = envir, globals = globals)
@@ -95,10 +96,19 @@ as_BatchtoolsFuture <- function(future,
                                 registry = list(),
                                 ...) {
   if (is.function(workers)) workers <- workers()
-  if (!is.null(workers)) {
+  if (is.null(workers)) {
+    workers <- getOption("future.batchtools.workers", default = 100)
+    stop_if_not(
+      is.numeric(workers),
+      length(workers) == 1,
+      !is.na(workers), workers >= 1
+    )
+  } else {
     stop_if_not(length(workers) >= 1)
     if (is.numeric(workers)) {
-      stop_if_not(!anyNA(workers), all(workers >= 1))
+      stop_if_not(length(workers) == 1, !is.na(workers), workers >= 1)
+    } else if (is.character(workers)) {
+      stop_if_not(length(workers) >= 0, !anyNA(workers))
     } else {
       stop("Argument 'workers' should be either a numeric or a function: ",
            mode(workers))
@@ -149,17 +159,24 @@ as_BatchtoolsFuture <- function(future,
 #'
 #' @export
 #' @keywords internal
-print.BatchtoolsFuture <- function(x, ...) {
+print.BatchtoolsFuture <- function(x, ...) {  
   NextMethod()
 
   ## batchtools specific
-  reg <- x$config$reg
+  config <- x$config
 
+  conf.file <- config$conf.file
+  printf("batchtools configuration file: %s\n", file_info(conf.file))
+  
+  reg <- config$reg
   if (inherits(reg, "Registry")) {
+    cluster.functions <- reg$cluster.functions
     printf("batchtools cluster functions: %s\n",
-           sQuote(reg$cluster.functions$name))
+           sQuote(cluster.functions$name))
+    template <- attr(cluster.functions, "template")
+    printf("batchtools cluster functions template: %s\n", file_info(template))
   } else {
-    printf("batchtools cluster functions: N/A\n")
+    printf("batchtools cluster functions: <none>\n")
   }
 
   ## Ask for status once
@@ -179,7 +196,7 @@ print.BatchtoolsFuture <- function(x, ...) {
       printf("  Work dir exists: %s\n", file_test("-d", reg$work.dir))
       try(print(reg))
     } else {
-      printf("batchtools Registry: N/A\n")
+      printf("batchtools Registry: <NA>\n")
     }
   }
 
@@ -207,9 +224,21 @@ status <- function(future, ...) {
       oopts <- list()
     }
     on.exit(options(oopts))
-    batchtools::getStatus(...)
+    ## WORKAROUND: batchtools::getStatus() updates the RNG state,
+    ## which we must make sure to undo.
+    with_stealth_rng({
+      batchtools::getStatus(...)
+    })
   } ## get_status()
 
+  ## Known to be in its final state?
+  if (getOption("future.batchtools.status.cache", TRUE)) {
+    status <- future$.status
+    if (identical(status, c("defined", "finished", "started", "submitted"))) {
+      return(status)
+    }
+  }
+
   config <- future$config
   reg <- config$reg
   if (!inherits(reg, "Registry")) return(NA_character_)
@@ -231,6 +260,9 @@ status <- function(future, ...) {
     if (result_has_errors(result)) status <- unique(c("error", status))
   }
 
+  ## Cache result
+  future$.status <- status
+  
   if (debug) mdebug("- status: ", paste(sQuote(status), collapse = ", "))
 
   status
@@ -315,14 +347,27 @@ loggedOutput.BatchtoolsFuture <- function(future, ...) {
 #' @export
 #' @keywords internal
 resolved.BatchtoolsFuture <- function(x, ...) {
-  ## Has internal future state already been switched to be resolved
-  resolved <- NextMethod()
-  if (resolved) return(TRUE)
+  signalEarly <- import_future("signalEarly")
+  
+  ## Is value already collected?
+  if (!is.null(x$result)) {
+    ## Signal conditions early?
+    signalEarly(x, ...)
+    return(TRUE)
+  }
+
+  ## Assert that the process that created the future is
+  ## also the one that evaluates/resolves/queries it.
+  assertOwner <- import_future("assertOwner")
+  assertOwner(x)
 
   ## If not, checks the batchtools registry status
   resolved <- finished(x)
   if (is.na(resolved)) return(FALSE)
- 
+
+  ## Signal conditions early? (happens only iff requested)
+  if (resolved) signalEarly(x, ...)
+
   resolved
 }
 
@@ -330,39 +375,61 @@ resolved.BatchtoolsFuture <- function(x, ...) {
 #' @export
 #' @keywords internal
 result.BatchtoolsFuture <- function(future, cleanup = TRUE, ...) {
+
+  debug <- getOption("future.debug", FALSE)
+  if (debug) {
+    mdebug("result() for BatchtoolsFuture ...")
+    on.exit(mdebug("result() for BatchtoolsFuture ... done"), add = TRUE)
+  }
+
   ## Has the value already been collected?
   result <- future$result
-  if (inherits(result, "FutureResult")) return(result)
+  if (inherits(result, "FutureResult")) {
+    if (debug) mdebug("- FutureResult already collected")
+    return(result)
+  }
 
   ## Has the value already been collected? - take two
   if (future$state %in% c("finished", "failed", "interrupted")) {
+    if (debug) mdebug("- FutureResult already collected - take 2")
     return(NextMethod())
   }
 
   if (future$state == "created") {
+    if (debug) mdebug("- starting future ...")
     future <- run(future)
+    if (debug) mdebug("- starting future ... done")
   }
 
+  if (debug) mdebug("- getting batchtools status")
   stat <- status(future)
   if (is_na(stat)) {
     label <- future$label
     if (is.null(label)) label <- "<none>"
-    stop(sprintf("The result no longer exists (or never existed) for Future ('%s') of class %s", label, paste(sQuote(class(future)), collapse = ", "))) #nolint
+    stopf("The result no longer exists (or never existed) for Future ('%s') of class %s", label, paste(sQuote(class(future)), collapse = ", ")) #nolint
   }
 
+  if (debug) mdebug("- waiting for batchtools job to finish ...")
   result <- await(future, cleanup = FALSE)
+  if (debug) mdebug("- waiting for batchtools job to finish ... done")
   stop_if_not(inherits(result, "FutureResult"))
   future$result <- result
   future$state <- "finished"
 
-  if (cleanup) delete(future)
+  if (cleanup) {
+    if (debug) mdebugf("- delete %s ...", class(future)[1])
+    delete(future)
+    if (debug) mdebugf("- delete %s ... done", class(future)[1])
+  }
 
+  if (debug) mdebug("- NextMethod()")
   NextMethod()
 }
 
 
 #' @importFrom future run getExpression
 #' @importFrom batchtools batchExport batchMap saveRegistry setJobNames submitJobs
+#' @importFrom utils capture.output str
 #' @export
 run.BatchtoolsFuture <- function(future, ...) {
   if (future$state != "created") {
@@ -448,16 +515,20 @@ run.BatchtoolsFuture <- function(future, ...) {
 
   ## 1. Add to batchtools for evaluation
   mdebug("batchtools::batchMap()")
-  jobid <- batchMap(fun = geval, list(expr),
-                    more.args = list(substitute = TRUE), reg = reg)
+  ## WORKAROUND: batchtools::batchMap() updates the RNG state,
+  ## which we must make sure to undo.
+  with_stealth_rng({
+    jobid <- batchMap(fun = geval, list(expr),
+                      more.args = list(substitute = TRUE), reg = reg)
+  })
 
-  ## 1b. Set job name, if specified
+  ## 2. Set job name, if specified
   label <- future$label
   if (!is.null(label)) {
     setJobNames(ids = jobid, names = label, reg = reg)
   }
   
-  ## 2. Update
+  ## 3. Update
   future$config$jobid <- jobid
   mdebugf("Created %s future #%d", class(future)[1], jobid$job.id)
 
@@ -483,15 +554,41 @@ run.BatchtoolsFuture <- function(future, ...) {
     }, error = function(ex) list())
   }
 
-  ## 3. Submit
+  ## 4. Wait for an available worker
+  waitForWorker(future, workers = future$workers)
+
+  ## 5. Submit
   future$state <- "running"
   resources <- future$config$resources
   if (is.null(resources)) resources <- list()
 
-  batchtools::submitJobs(reg = reg, ids = jobid, resources = resources)
+  ## WORKAROUND: batchtools::submitJobs() updates the RNG state,
+  ## which we must make sure to undo.
+  tryCatch({
+    with_stealth_rng({
+      submitJobs(reg = reg, ids = jobid, resources = resources)
+    })
+  }, error = function(ex) {
+    msg <- conditionMessage(ex)
+    label <- future$label
+    if (is.null(label)) label <- "<none>"
+    msg <- sprintf("Failed to submit %s (%s). The reason was: %s", class(future)[1], label, msg)
+    info <- capture.output(str(resources))
+    info <- paste(info, collapse = "\n")
+    msg <- sprintf("%s\nTROUBLESHOOTING INFORMATION:\nbatchtools::submitJobs() was called with the following 'resources' argument:\n%s\n", msg, info)
+    stop(BatchtoolsFutureError(msg, future = future))
+  })
 
   mdebugf("Launched future #%d", jobid$job.id)
 
+  ## 6. Rerserve worker for future
+  registerFuture(future)
+
+  ## 7. Trigger early signalling
+  if (inherits(future, "BatchtoolsUniprocessFuture")) {
+    resolved(future)
+  }
+  
   invisible(future)
 } ## run()
 
@@ -507,6 +604,7 @@ await <- function(future, cleanup = TRUE,
   stop_if_not(is.finite(alpha), alpha > 0)
   
   debug <- getOption("future.debug", FALSE)
+  if (debug) mdebug("future.batchtools:::await() ...")
 
   expr <- future$expr
   config <- future$config
@@ -525,10 +623,12 @@ await <- function(future, cleanup = TRUE,
  
   res <- waitForJobs(ids = jobid, timeout = timeout, sleep = sleep_fcn,
                      stop.on.error = FALSE, reg = reg)
-  mdebugf("- batchtools::waitForJobs(): %s", res)
+  if (debug) mdebugf("- batchtools::waitForJobs(): %s", res)
   stat <- status(future)
-  mdebugf("- status(): %s", paste(sQuote(stat), collapse = ", "))
-  mdebug("batchtools::waitForJobs() ... done")
+  if (debug) {
+    mdebugf("- status(): %s", paste(sQuote(stat), collapse = ", "))
+    mdebug("batchtools::waitForJobs() ... done")
+  }
 
   finished <- is_na(stat) || any(c("finished", "error", "expired") %in% stat)
 
@@ -541,12 +641,23 @@ await <- function(future, cleanup = TRUE,
     label <- future$label
     if (is.null(label)) label <- "<none>"
     if ("finished" %in% stat) {
+      if (debug) mdebug("- batchtools::loadResult() ...")
       result <- loadResult(reg = reg, id = jobid)
+      if (debug) mdebug("- batchtools::loadResult() ... done")
+      
       if (inherits(result, "FutureResult")) {
         prototype_fields <- c(prototype_fields, "batchtools_log")
-        result[["batchtools_log"]] <- try({
+        result[["batchtools_log"]] <- try(local({
+          if (debug) {
+            mdebug("- batchtools::getLog() ...")
+            on.exit(mdebug("- batchtools::getLog() ... done"))
+          }
+	  ## Since we're already collected the results, the log file
+	  ## should already exist, if it exists.  Because of this,
+	  ## only poll for the log file for a second before giving up.
+	  reg$cluster.functions$fs.latency <- 1.0
           getLog(id = jobid, reg = reg)
-        }, silent = TRUE)
+        }), silent = TRUE)
         if (result_has_errors(result)) cleanup <- FALSE
       }
     } else if ("error" %in% stat) {
@@ -574,7 +685,7 @@ await <- function(future, cleanup = TRUE,
       msg <- sprintf("BatchtoolsDeleted: Cannot retrieve value. Future ('%s') deleted: %s", label, reg$file.dir) #nolint
       stop(BatchtoolsFutureError(msg, future = future))
     }
-    if (debug) { mstr(result) }
+    if (debug) { mstr(result) }    
   } else {
     cleanup <- FALSE
     msg <- sprintf("AsyncNotReadyError: Polled for results for %s seconds every %g seconds, but asynchronous evaluation for future ('%s') is still running: %s", timeout, delta, label, reg$file.dir) #nolint
@@ -590,6 +701,8 @@ await <- function(future, cleanup = TRUE,
     delete(future, delta = 0.5 * delta, ...)
   }
 
+  if (debug) mdebug("future.batchtools:::await() ... done")
+
   result
 } # await()
 
@@ -673,6 +786,9 @@ delete.BatchtoolsFuture <- function(future,
   result <- result(future, cleanup = FALSE)
   stop_if_not(inherits(result, "FutureResult"))
 
+  ## Free up worker
+  unregisterFuture(future)
+
   ## To simplify post mortem troubleshooting in non-interactive sessions,
   ## should the batchtools registry files be removed or not?
   mdebugf("delete(): Option 'future.delete = %s",
@@ -705,15 +821,23 @@ delete.BatchtoolsFuture <- function(future,
   on.exit(options(oopts))
 
   ## Try to delete registry
-  interval <- delta
-  for (kk in seq_len(times)) {
-    try(clearRegistry(reg = reg), silent = TRUE)
-    try(removeRegistry(wait = 0.0, reg = reg), silent = FALSE)
-    if (!file_test("-d", path)) break
-    Sys.sleep(interval)
-    interval <- alpha * interval
-  }
-
+  ## WORKAROUND: batchtools::clearRegistry() and
+  ## batchtools::removeRegistry() update the RNG state,
+  ## which we must make sure to undo.
+  with_stealth_rng({
+    interval <- delta
+    for (kk in seq_len(times)) {
+      try(unlink(path, recursive = TRUE), silent = FALSE)
+      if (!file_test("-d", path)) break
+      try(removeRegistry(wait = 0.0, reg = reg), silent = FALSE)
+      if (!file_test("-d", path)) break
+      try(clearRegistry(reg = reg), silent = TRUE)
+      try(removeRegistry(wait = 0.0, reg = reg), silent = FALSE)
+      if (!file_test("-d", path)) break
+      Sys.sleep(interval)
+      interval <- alpha * interval
+    }
+  })
 
   ## Success?
   if (file_test("-d", path)) {
