@@ -14,7 +14,7 @@
 #' @param \ldots Additional arguments passed
 #' to [BatchtoolsFuture()].
 #'
-#' @return An object of class `BatchtoolsFuture`.
+#' @return An object of class `BatchtoolsMulticoreFuture`.
 #'
 #' @details
 #' batchtools multicore futures rely on the batchtools backend set
@@ -23,7 +23,7 @@
 #' supporting the `ps` command-line tool, e.g. Linux and macOS.
 #'
 #' @importFrom batchtools makeClusterFunctionsMulticore
-#' @importFrom future availableCores
+#' @importFrom parallelly availableCores
 #' @export
 #' @keywords internal
 batchtools_multicore <- function(expr, envir = parent.frame(),
@@ -38,7 +38,8 @@ batchtools_multicore <- function(expr, envir = parent.frame(),
             is.finite(workers), workers >= 1L)
 
   ## Fall back to batchtools_local if multicore processing is not supported
-  if (workers == 1L || is_os("windows") || is_os("solaris") ||
+  if ((workers == 1L && !inherits(workers, "AsIs")) ||
+      is_os("windows") || is_os("solaris") ||
       availableCores(constraints = "multicore") == 1L) {
     ## covr: skip=1
     return(batchtools_local(expr, envir = envir, substitute = FALSE,
@@ -51,16 +52,21 @@ batchtools_multicore <- function(expr, envir = parent.frame(),
 
   cf <- makeClusterFunctionsMulticore(ncpus = workers)
 
-  future <- BatchtoolsFuture(expr = expr, envir = envir, substitute = FALSE,
-                            globals = globals,
-                            label = label,
-                            cluster.functions = cf,
-                            registry = registry, 
-                            ...)
+  future <- BatchtoolsMulticoreFuture(
+    expr = expr, envir = envir, substitute = FALSE,
+    globals = globals,
+    label = label,
+    cluster.functions = cf,
+    registry = registry, 
+    ...
+  )
 
   if (!future$lazy) future <- run(future)
 
-  future
+  invisible(future)
 }
-class(batchtools_multicore) <- c("batchtools_multicore", "batchtools",
-                                 "multiprocess", "future", "function")
+class(batchtools_multicore) <- c(
+  "batchtools_multicore", "batchtools_multiprocess", "batchtools",
+  "multiprocess", "future", "function"
+)
+attr(batchtools_multicore, "tweakable") <- c("finalize")
